#  ExoIris: fast, flexible, and easy exoplanet transmission spectroscopy in Python.
#  Copyright (C) 2024 Hannu Parviainen
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path

from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.pyplot import subplots, setp
from numpy import sqrt, array
from pytransit import LDTkLD as PTLDTkLD
from ldtk import BoxcarFilter

from .tsdata import TSData


class LDTkLD(PTLDTkLD):
    def __init__(self, data: TSData,
                 teff: tuple[float, float],
                 logg: tuple[float, float],
                 metal: tuple[float, float],
                 cache: str | Path | None = None,
                 dataset: str = 'vis-lowres') -> None:
        filters = [BoxcarFilter('a', wla*1e3, wlb*1e3) for wla, wlb in zip(data._wl_l_edges, data._wl_r_edges)]
        super().__init__(filters, teff, logg, metal, cache, dataset)
        self.wavelength = data.wavelength
        self.dataset = dataset

    def plot_profiles(self, teff: float, logg: float, metal: float, x: str = 'mu', ax: Axes = None) -> Figure:
        """Plots the profiles of a star's surface brightness.

        Parameters
        ----------
        teff : float
            Effective temperature of the star.

        logg : float
            Logarithm of the surface gravity of the star.

        metal : float
            Metallicity of the star.

        x : str, optional
            The x-axis coordinate to plot against. Default is 'mu'.
            Must be either 'mu' or 'z'.

        ax : Axes, optional
            The matplotlib axes to plot the profiles on. If not provided,
            a new figure and axes will be created.

        Returns
        -------
        Figure
            The matplotlib figure containing the plot.

        Raises
        ------
        ValueError
            If the provided value of x is not 'mu' or 'z'.
        """
        if ax is None:
            fig, ax = subplots()
        else:
            fig = ax.get_figure()

        if x == 'z':
            xv = sqrt(1-self.mu**2)
            xl = 'z [R$_\star$]'
        elif x == 'mu':
            xv = self.mu
            xl = '$\mu$'
        else:
            raise ValueError("x must be either 'mu' or 'z'")

        ldi, _ = self(self.mu, array([teff, logg, metal]))
        l = ax.pcolormesh(xv, self.wavelength, ldi[0])
        fig.colorbar(l, label='Surface brightness')
        setp(ax, xlabel=xl, ylabel='Wavelength  [$\mu$m]' )
        fig.tight_layout()
        return fig



#  ExoIris: fast, flexible, and easy exoplanet transmission spectroscopy in Python.
#  Copyright (C) 2024 Hannu Parviainen
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

from matplotlib.figure import Figure
from matplotlib.pyplot import subplots, setp
from numpy import array, log10, diff, sqrt, floor, ceil, arange, newaxis
from scipy.optimize import minimize

from pytransit import BaseLPF, LinearModelBaseline
from pytransit.orbits import as_from_rhop, i_from_ba, fold, i_from_baew, d_from_pkaiews, epoch

from .tslpf import TSLPF

class WhiteLPF(BaseLPF):
    def __init__(self, tsa: TSLPF):
        times = tsa.time
        fluxes = [f.mean(0) for f in tsa.flux]
        covs = [(t-t.mean())[:, newaxis] for t in times]
        wnids = arange(len(fluxes))

        super().__init__('white', ['white'], times, fluxes, covariates=covs, wnids=wnids)
        self.set_prior('tc', tsa.ps[tsa.ps.find_pid('tc')].prior)
        self.set_prior('p', tsa.ps[tsa.ps.find_pid('p')].prior)
        self.set_prior('rho', tsa.ps[tsa.ps.find_pid('rho')].prior)
        self.set_prior('b', tsa.ps[tsa.ps.find_pid('b')].prior)
        pr = tsa.ps[tsa._sl_rratios][0].prior
        if 'Uniform' in str(pr.__class__):
            self.set_prior('k2', 'UP', pr.a**2, pr.b**2)
        if 'Normal' in str(pr.__class__):
            self.set_prior('k2', 'NP', pr.mean**2, pr.std**2)
        self.set_prior('wn_loge_0', 'NP', log10(diff(self.ofluxa).std() / sqrt(2)), 1e-5)

    def _init_baseline(self):
        self._add_baseline_model(LinearModelBaseline(self))

    def optimize(self, pv0=None, method='powell', maxfev: int = 5000):
            if pv0 is None:
                if self.de is not None:
                    pv0 = self.de.minimum_location
                else:
                    pv0 = self.ps.mean_pv
            res = minimize(lambda pv: -self.lnposterior(pv), pv0, method=method, options={'maxfev':maxfev})
            self._local_minimization = res

    @property
    def transit_center(self):
        pv = self._local_minimization.x
        return pv[0] + pv[1]*epoch(self.times[0].mean(), pv[0], pv[1])

    @property
    def transit_duration(self):
        pv = self._local_minimization.x
        a = as_from_rhop(pv[2], pv[1])
        i = i_from_ba(pv[3], a)
        t14 = d_from_pkaiews(pv[1], sqrt(pv[4]), a, i, 0., 0., 1, 14)
        return t14

    def plot(self, axs=None, figsize=None, ncols=2) -> Figure:
        if axs is None:
            nrows = int(ceil(self.nlc / ncols))
            fig, axs = subplots(nrows, ncols, figsize=figsize, sharey='all', squeeze=False)
        else:
            fig = axs[0].get_figure()

        tref = floor(self.timea.min())
        fm = self.flux_model(self._local_minimization.x)
        t14 = self.transit_duration
        pv = self._local_minimization.x

        for i, sl in enumerate(self.lcslices):
            ax = axs.flat[i]
            tc = pv[0] + pv[1]*epoch(self.times[i].mean(), pv[0], pv[1])
            ax.plot(self.timea[sl] - tref, self.ofluxa[sl], '.k', alpha=0.25)
            ax.plot(self.timea[sl] - tref, fm[sl], 'k')
            ax.axvline(tc - tref, ls='--', c='0.5')
            ax.axvline(tc - tref - 0.5*t14, ls='--', c='0.5')
            ax.axvline(tc - tref + 0.5*t14, ls='--', c='0.5')
            setp(ax, xlabel=f'Time - {tref:.0f} [BJD]', xlim=(self.times[i].min()-tref, self.times[i].max()-tref))
        setp(axs[:,0], ylabel='Normalized flux')
        return fig



#  ExoIris: fast, flexible, and easy exoplanet transmission spectroscopy in Python.
#  Copyright (C) 2024 Hannu Parviainen
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import codecs
import json
import pickle
from multiprocessing import Pool
from pathlib import Path
from collections.abc import Sequence
from typing import Optional, Callable, Any

import pandas as pd
import seaborn as sb
import astropy.io.fits as pf
import astropy.units as u

from astropy.table import Table
from celerite2 import GaussianProcess, terms
from matplotlib.pyplot import subplots, setp, figure, Figure, GridSpec, Axes
from numpy import (poly1d, polyfit, where, sqrt, clip, percentile, median, squeeze, floor, ndarray,
                   array, inf, newaxis, r_, arange, tile, log10, sort, argsort, concatenate)
from numpy.random import normal, permutation
from pytransit.orbits import fold, epoch
from pytransit.param import ParameterSet
from pytransit.utils.de import DiffEvol
from scipy.stats import norm
from uncertainties import ufloat, UFloat

from .ldtkld import LDTkLD
from .tsdata import TSData, TSDataSet
from .tslpf import TSLPF
from .wlpf import WhiteLPF


def load_model(fname: Path | str, name: str | None = None):
    """Load an ExoIris analysis from a FITS file.

    Parameters
    ----------
    fname
        The name of the savefile.
    name
        The name of the new ExoIris model. If not provided, the original analysis name will be used.

    Returns
    -------
    ExoIris
        The saved ExoIris analysis.

    Raises
    ------
    IOError
        If there is an error while opening or reading the file.

    ValueError
        If the file format is invalid or does not match the expected format.
    """
    with pf.open(fname) as hdul:
        d = []
        for i in range(hdul[0].header['NDGROUPS']):
            d.append(TSData(hdul[f'TIME_{i}'].data.astype('d'), hdul[f'WAVELENGTH_{i}'].data.astype('d'),
                            hdul[f'FLUX_{i}'].data.astype('d'), hdul[f'FERR_{i}'].data.astype('d'),
                            name=hdul[f'FLUX_{i}'].header['NAME']))
        data = TSDataSet(d)

        if hdul[0].header['LDMODEL'] == 'ldtk':
            filters, teff, logg, metal, dataset = pickle.loads(codecs.decode(json.loads(hdul[0].header['LDTKLD']).encode(), "base64"))
            ldm = LDTkLD(filters, teff, logg, metal, dataset=dataset)
        else:
            ldm =  hdul[0].header['LDMODEL']

        a = ExoIris(name or hdul[0].header['NAME'], ldmodel=ldm, data=data)
        a.set_radius_ratio_knots(hdul['K_KNOTS'].data.astype('d'))
        a.set_limb_darkening_knots(hdul['LD_KNOTS'].data.astype('d'))
        a.transit_center = hdul[0].header['T0']
        a.transit_duration = hdul[0].header['T14']

        if a.transit_duration is not None:
            a.ootmask = abs(a.time - a.transit_center) > 0.502 * a.transit_duration

        priors = pickle.loads(codecs.decode(json.loads(hdul['PRIORS'].header['PRIORS']).encode(), "base64"))
        a._tsa.ps = ParameterSet([pickle.loads(p) for p in priors])
        a._tsa.ps.freeze()
        if 'DE' in hdul:
            a._tsa._de_population = Table(hdul['DE'].data).to_pandas().values
            a._tsa._de_imin = hdul['DE'].header['IMIN']
        if 'MCMC' in hdul:
            npop = hdul['MCMC'].header['NPOP']
            ndim = hdul['MCMC'].header['NDIM']
            a._tsa._mc_chains = Table(hdul['MCMC'].data).to_pandas().values.reshape([npop, -1, ndim])
        return a


class ExoIris:
    """The core ExoIris class providing tools for exoplanet transit spectroscopy.

    Parameters
    ----------
    name
        The name of the instance.
    ldmodel
        The model for the limb darkening.
    data
        The time-series data object.
    nk
        The number of kernel samples.
    nldc
        The number of limb darkening coefficients.
    nthreads
        The number of threads to use for computation.
    tmpars
        Additional transit model parameters.
    noise_model
        The noise model to use. Should be either "white" for white noise or "fixed_gp" for Gaussian Process.
    """

    def __init__(self, name: str, ldmodel, data: TSDataSet | TSData, nk: int = None, nldc: int = 10, nthreads: int = 1,
                 tmpars: dict | None = None, noise_model: str = 'white'):
        data = TSDataSet([data]) if isinstance(data, TSData) else data
        self._tsa = TSLPF(name, ldmodel, data, nk=nk, nldc=nldc, nthreads=nthreads, tmpars=tmpars, noise_model=noise_model)
        self._wa = None
        self.nthreads = nthreads
        self.de: DiffEvol | None = None
        self.sampler = None

        self.transit_center: float | None = None
        self.transit_duration: float | None= None

        self._tref = floor(concatenate(self.time).min())

    def lnposterior(self, pvp: ndarray) -> ndarray:
        """Calculate the log posterior probability for a single parameter vector or an array of parameter vectors.

        Parameters
        ----------
        pvp
            The vector of parameter values or an array of parameter vectors with a shape [npv, np].

        Returns
        -------
        ndarray
            The natural logarithm of the posterior probability.
        """
        return squeeze(self._tsa.lnposterior(pvp))

    def set_noise_model(self, noise_model: str) -> None:
        """Set the noise model for the analysis.

        Parameters
        ----------
        noise_model
            The noise model to be used. Must be one of the following: white, fixed_gp, free_gp.

        Raises
        ------
        ValueError
            If noise_model is not one of the specified options.
        """
        self._tsa.set_noise_model(noise_model)

    def set_data(self, data: TSData | TSDataSet) -> None:
        """Set the model data.

        Parameters
        ----------
        data
           The spectroscopic transit light curve.
        """
        data = TSDataSet([data]) if isinstance(data, TSData) else data
        self._tsa.set_data(data)

    def set_prior(self, parameter: str, prior: str | Any, *nargs) -> None:
        """Set a prior on a model parameter.

        Parameters
        ----------
        parameter
            The name of the parameter to set prior for.

        prior
            The prior distribution to set for the parameter. This can be "NP" for a normal prior, "UP" for a
            uniform prior, or an object with .logpdf(x) method.

        *nargs
            Additional arguments to be passed to the prior: (mean, std) for the normal prior and (min, max)
            for the uniform prior.
        """
        self._tsa.set_prior(parameter, prior, *nargs)

    def set_radius_ratio_prior(self, prior: float, *nargs) -> None:
        """Set an identical prior on all radius ratio (k) knots.

        Parameters
        ----------
        prior
            The prior for the radius ratios.
        *nargs
            Additional arguments for the prior.
        """
        for l in self._tsa.k_knots:
            self.set_prior(f'k_{l:08.5f}', prior, *nargs)

    def set_ldtk_prior(self,
                       teff: UFloat | tuple[float, float],
                       logg: UFloat | tuple[float, float],
                       metal: UFloat | tuple[float, float],
                       dataset: str = 'visir-lowres', width: float = 50.0, uncertainty_multiplier: float = 10.0):
        """Set priors on the limb darkening parameters using LDTk.

        Sets priors on the limb darkening parameters based on theoretical stellar models using LDTk.

        Parameters
        ----------
        teff
            The effective temperature and its uncertainty in Kelvin.
        logg
            The surface gravity and its uncertainty in cm/s^2.
        metal
            The metallicity and its uncertainty.
        dataset
            The name of the dataset. Default is 'visir-lowres'.
        width
            The passband width in nanometers. Default is 50.
        uncertainty_multiplier
            The uncertainty multiplier to adjust the width of the prior. Default is 10.

        """
        teff = (teff.n, teff.s) if isinstance(teff, UFloat) else teff
        logg = (logg.n, logg.s) if isinstance(logg, UFloat) else logg
        metal = (metal.n, metal.s) if isinstance(metal, UFloat) else metal
        self._tsa.set_ldtk_prior(teff, logg, metal, dataset, width, uncertainty_multiplier)

    def set_gp_hyperparameters(self, sigma: float, rho: float) -> None:
        """Set Gaussian Process (GP) hyperparameters assuming a Matern-3/2 kernel.

        Parameters
        ----------
        sigma
            The kernel amplitude parameter.
        rho
            The length scale parameter.
        """
        self._tsa.set_gp_hyperparameters(sigma, rho)

    def set_gp_kernel(self, kernel: terms.Term) -> None:
        """Set the Gaussian Process (GP) kernel.

        Parameters
        ----------
        kernel
            The kernel to set for the GP.
        """
        self._tsa.set_gp_kernel(kernel)

    @property
    def name(self) -> str:
        """Get the name of the analysis."""
        return self._tsa.name

    @name.setter
    def name(self, name: str):
        self._tsa.name = name

    @property
    def data(self) -> TSDataSet:
        """Get the analysis dataset."""
        return self._tsa.data

    @property
    def time(self) -> ndarray:
        """Get the concatenated time array."""
        return self._tsa.time

    @property
    def wavelength(self) -> ndarray:
        """Get the concatenated wavelength array."""
        return self._tsa.wavelength

    @property
    def fluxes(self) -> ndarray:
        """Get the concatenated flux array."""
        return self._tsa.flux

    @property
    def errors(self) -> ndarray:
        """Get the concatenated flux uncertainty array."""
        return self._tsa.ferr

    @property
    def k_knots(self) -> ndarray:
        """Get the radius ratio (k) knots."""
        return self._tsa.k_knots

    @property
    def ndim(self) -> int:
        """Get the number of free model parameters."""
        return self._tsa.ndim

    @property
    def nk(self) -> int:
        """Get the number of radius ratio (k) knots."""
        return self._tsa.nk

    @property
    def nldp(self) -> int:
        """Get the number of limb darkening knots."""
        return self._tsa.nldc

    @property
    def npb(self) -> int:
        """Get the number of passbands."""
        return self._tsa.npb

    @property
    def ldmodel(self):
        """Get the limb darkening model."""
        return self._tsa.ldmodel

    @property
    def gp(self) -> GaussianProcess:
        """Get the Gaussian Process (GP) model."""
        return self._tsa._gp

    @property
    def optimizer_population(self) -> ndarray:
        """Get the current population of the optimizer."""
        return self._tsa._de_population

    @property
    def mcmc_chains(self) -> ndarray:
        """Get the current emcee MCMC chains."""
        return self._tsa._mc_chains

    @property
    def posterior_samples(self) -> pd.DataFrame:
        """Get the posterior samples from the MCMC sampler."""
        return pd.DataFrame(self._tsa._mc_chains.reshape([-1, self.ndim]), columns=self.ps.names)

    def add_radius_ratio_knots(self, knot_wavelengths: Sequence) -> None:
        """Add radius ratio (k) knots.

        Parameters
        ----------
        knot_wavelengths
            List or array of knot wavelengths to be added.
        """
        self._tsa.add_k_knots(knot_wavelengths)

    def set_radius_ratio_knots(self, knot_wavelengths: Sequence) -> None:
        """Set the radius ratio (k) knots.

        Parameters
        ----------
        knot_wavelengths
            List or array of knot wavelengths.
        """
        self._tsa.set_k_knots(knot_wavelengths)

    def add_limb_darkening_knots(self, knot_wavelengths: Sequence) -> None:
        """Add limb darkening knots.

        Parameters
        ----------
        knot_wavelengths
            List or array of knot wavelengths to be added.
        """
        raise NotImplementedError()
        self._tsa.add_limb_darkening_knots(knot_wavelengths)

    def set_limb_darkening_knots(self, knot_wavelengths: Sequence) -> None:
        """Set the limb darkening knots.

        Parameters
        ----------
        knot_wavelengths
            List or array of knot wavelengths.
        """
        self._tsa.set_ld_knots(knot_wavelengths)

    @property
    def ps(self) -> ParameterSet:
        """Get the model parameterization."""
        return self._tsa.ps

    def print_parameters(self) -> None:
        """Print the model parameterization."""
        self._tsa.print_parameters(1)

    def plot_setup(self, figsize: tuple[float, float] | None =None, xscale: str | None = None, xticks: Sequence | None = None) -> Figure:
        """Plot the model setup with limb darkening knots, radius ratio knots, and data binning.

        Parameters
        ----------
        figsize
            The size of the figure in inches.
        xscale
            The scale of the x-axis. If provided, the x-axis scale of all the subplots will be set to this value.
        xticks
            The list of x-axis tick values for all the subplots. If provided, the x-axis ticks of all the subplots will
            be set to these values.

        Returns
        -------
        Figure
            The matplotlib Figure object that contains the created subplots.

        """
        using_ldtk = isinstance(self._tsa.ldmodel, LDTkLD)

        if not using_ldtk:
            figsize = figsize or (13, 4)
            fig, axs = subplots(3, 1, figsize=figsize, sharex='all', sharey='all')
            axl, axk, axw = axs

            axl.vlines(self._tsa.ld_knots, 0.1, 0.5, ec='k')
            axl.text(0.01, 0.90, 'Limb darkening knots', va='top', transform=axl.transAxes)
        else:
            figsize = figsize or (13, 2*4/3)
            fig, axs = subplots(2, 1, figsize=figsize, sharex='all', sharey='all')
            axk, axw = axs
            axl = None

        axk.vlines(self._tsa.k_knots, 0.1, 0.5, ec='k')
        axk.text(0.01, 0.90, 'Radius ratio knots', va='top', transform=axk.transAxes)
        for ds in self.data:
            axw.vlines(ds.wavelength, 0.1, 0.5, ec='k')
        axw.text(0.01, 0.90, 'Wavelength bins', va='top', transform=axw.transAxes)

        if not using_ldtk:
            sb.despine(ax=axl, top=False, bottom=True, right=False)
            sb.despine(ax=axk, top=True, bottom=True, right=False)
        else:
            sb.despine(ax=axk, top=False, bottom=True, right=False)

        sb.despine(ax=axw, top=True, bottom=False, right=False)
        setp(axs, xlim=(self.data.wlmin-0.02, self.data.wlmax+0.02), yticks=[], ylim=(0, 0.9))
        setp(axw, xlabel=r'Wavelength [$\mu$m]')
        setp(axs[0].get_xticklines(), visible=False)
        setp(axs[0].get_xticklabels(), visible=False)
        setp(axs[1].get_xticklines(), visible=False)
        setp(axs[-1].get_xticklines(), visible=True)

        if xscale:
            setp(axs, xscale=xscale)
        if xticks is not None:
            [ax.set_xticks(xticks, labels=xticks) for ax in axs]
        fig.tight_layout()
        return fig

    def fit_white(self) -> None:
        """Fit a white light curve model and sets the out-of-transit mask."""
        self._wa = WhiteLPF(self._tsa)
        self._wa.optimize()
        pv = self._wa._local_minimization.x
        self.transit_center = self._wa.transit_center
        self.transit_duration = self._wa.transit_duration
        self.data.calculate_ootmask(pv[0], pv[1], self.transit_duration)

    def plot_white(self, axs=None, figsize=None, ncols=2) -> Figure:
        """Plot the white light curve data with the best-fit model.

        Parameters
        ----------
        axs : Axes, optional
            Matplotlib axis object on which to plot. If None, a new figure and axis will be created.
        figsize : tuple of float, optional
            Tuple representing the figure size in inches. Default is None.
        ncols : int, optional
            Number of columns in the plot layout. Default is 2.
        """
        return self._wa.plot(axs=axs, figsize=figsize, ncols=ncols)

    def normalize_baseline(self, deg: int = 1) -> None:
        """Normalize the baseline flux for each spectroscopic light curve.

        Normalize the baseline flux using a low-order polynomial fitted to the out-of-transit
        data for each spectroscopic light curve.

        Parameters
        ----------
        deg
            The degree of the fitted polynomial. Should be 0 or 1. Higher degrees are not allowed
            because they could affect the transit depths.

        Raises
        ------
        ValueError
            If `deg` is greater than 1.

        Notes
        -----
        This method normalizes the baseline of the fluxes for each planet. It fits a polynomial of degree
        `deg` to the out-of-transit data points and divides the fluxes by the fitted polynomial evaluated
        at each time point.
        """
        if deg > 1:
            raise ValueError("The degree of the fitted polynomial ('deg') should be 0 or 1. Higher degrees are not allowed because they could affect the transit depths.")
        for d in self.data:
            for ipb in range(d.nwl):
                pl = poly1d(polyfit(d.time[d.ootmask], d.fluxes[ipb, d.ootmask], deg=deg))(d.time)
                d.fluxes[ipb, :] /= pl
                d.errors[ipb, :] /= pl

    def plot_baseline(self, axs: Optional[Sequence[Axes]] = None, figsize=None) -> Figure:
        """Plot the out-of-transit spectroscopic light curves before and after the normalization.

        Parameters
        ----------
        axs
            Array of axes to plot on. If None, new axes will be created.

        Returns
        -------
        Figure
            The figure containing the subplots.
        """
        if axs is None:
            fig, axs = subplots(self.data.ngroups, 2, figsize=figsize, squeeze=False, constrained_layout=True)
        else:
            fig = axs[0,0].figure

        for i in range(self.data.ngroups):
            self._tsa._original_data[i].plot(ax=axs[i, 0], data=where(self.data[i].ootmask, self._tsa._original_data[i].fluxes, 1))
            self.data[i].plot(ax=axs[i, 1], data=where(self.data[i].ootmask, self.data[i].fluxes, 1))
        return fig

    def fit(self, niter: int = 200, npop: Optional[int] = None, pool: Optional[Pool] = None, lnpost: Optional[Callable]=None,
            population: Optional[ndarray] = None, initial_population: Optional[ndarray] = None) -> None:
        """Fit the spectroscopic light curves jointly using Differential Evolution.

        Fit the spectroscopic light curves jointly for `niter` iterations using Differential Evolution.

        Parameters
        ----------
        niter
            Number of iterations for optimization. Default is 200.
        npop
            Population size for optimization. Default is 150.
        pool
            Multiprocessing pool for parallel optimization. Default is None.
        lnpost
            Log posterior function for optimization. Default is None.
        """
        if population is not None:
            x0 = population
            npop = x0.shape[0]
        else:
            if self._tsa.de is None and initial_population is not None:
                x0 = initial_population
                npop = x0.shape[0]
            elif self._tsa._de_population is not None:
                x0 = self._tsa._de_population
                npop = x0.shape[0]
            else:
                if npop is None:
                    raise ValueError("'npop' cannot be None when starting global optimization from the white light curve fit.'")
                if npop <= 2*self._tsa.ndim:
                    raise ValueError("'npop' should be at least two times the number of free model parameters.")

                pv0 = self._wa._local_minimization.x
                x0 = self._tsa.ps.sample_from_prior(npop)
                x0[:, 0] = normal(pv0[2], 0.05, size=npop)
                x0[:, 1] = normal(pv0[0], 1e-4, size=npop)
                x0[:, 2] = normal(pv0[1], 1e-5, size=npop)
                x0[:, 3] = clip(normal(pv0[3], 0.01, size=npop), 0.0, 1.0)
                x0[:, self._tsa._sl_rratios] = normal(sqrt(pv0[4]), 0.001, size=(npop, self.nk))

        self._tsa.optimize_global(niter=niter, npop=npop, population=x0, pool=pool, lnpost=lnpost,
                                  vectorize=(pool is None))
        self.de = self._tsa.de

    def sample(self, niter: int = 500, thin: int = 10, repeats: int = 1, pool=None, lnpost=None, leave=True, save=False, use_tqdm: bool = True):
        """Sample the posterior distribution using the emcee MCMC sampler.

        Parameters
        ----------
        niter
            Number of iterations in the MCMC sampling. Default is 500.
        thin
            Thinning factor for the MCMC samples. Default is 10.
        repeats
            Number of repeated iterations in the MCMC sampling. Default is 1.
        pool
            Parallel processing pool object to use for parallelization. Default is None.
        lnpost
            Log posterior function that takes a parameter vector as input and returns the log posterior probability.
            Default is None.
        leave
            Whether to leave the progress bar visible after sampling is finished. Default is True.
        save
            Whether to save the MCMC samples to disk. Default is False.
        use_tqdm
            Whether to use tqdm progress bar during sampling. Default is True.

        """
        self._tsa.sample_mcmc(niter=niter, thin=thin, repeats=repeats, pool=pool, lnpost=lnpost,
                              vectorize=(pool is None), leave=leave, save=save, use_tqdm=use_tqdm)
        self.sampler = self._tsa.sampler

    def reset_sampler(self) -> None:
        """Reset the MCMC sampler

        Reset the MCMC sampler by clearing the Monte Carlo chains and setting the sampler to None.
        """
        self._tsa._mc_chains = None
        self._tsa.sampler = None

    def plot_transmission_spectrum(self, result: Optional[str] = None, ax: Axes = None, xscale: Optional[str] = None,
                                   xticks=None, ylim=None,  plot_resolution: bool = True) -> Figure:
        """Plot the transmission spectrum.

        Parameters
        ----------
        result
            The type of result to plot. Can be 'fit', 'mcmc', or None. If None, the default behavior is to use 'mcmc' if
            the MCMC sampler has been run, otherwise 'fit'. Default is None.
        ax
            The matplotlib Axes object to plot on. If None, a new figure and axes will be created. Default is None.
        xscale
            The scale of the x-axis. Can be 'linear', 'log', 'symlog', 'logit', or None. If None, the default behavior is to
            use the scale of the current axes. Default is None.
        xticks
            The tick locations for the x-axis. If None, the default behavior is to use the tick locations of the current axes.
        ylim
            The limits for the y-axis. If None, the default behavior is to use the limits of the current axes.
        plot_resolution
            Whether to plot the resolution of the transmission spectrum as vertical lines. Default is True.

        Returns
        -------
        Figure
            The matplotlib Figure of the plotted transmission spectrum.

        """
        if result is None:
            result = 'mcmc' if self._tsa.sampler is not None else 'fit'
        if result not in ('fit', 'mcmc'):
            raise ValueError("Result must be either 'fit', 'mcmc', or None")
        if result == 'mcmc' and self._tsa._mc_chains is None:
            raise ValueError("Cannot plot posterior solution before running the MCMC sampler.")

        fig, ax = subplots() if ax is None else (ax.get_figure(), ax)

        wavelength = concatenate(self.wavelength)
        ix = argsort(wavelength)

        if result == 'fit':
            pv = self._tsa._de_population[self._tsa._de_imin]
            ks = self._tsa._eval_k(pv[self._tsa._sl_rratios])
            ar = 1e2 * concatenate([squeeze(k) for k in ks]) ** 2
            ax.plot(wavelength[ix], ar[ix], c='k')
            ax.plot(self._tsa.k_knots, 1e2 * pv[self._tsa._sl_rratios] ** 2, 'k.')
        else:
            df = pd.DataFrame(self._tsa._mc_chains.reshape([-1, self._tsa.ndim]), columns=self._tsa.ps.names)
            ks = self._tsa._eval_k(df.iloc[:, self._tsa._sl_rratios])
            ar = 1e2 * concatenate(ks, axis=1) ** 2
            ax.fill_between(wavelength[ix], *percentile(ar[:, ix], [16, 84], axis=0), alpha=0.25)
            ax.plot(wavelength[ix], median(ar, 0)[ix], c='k')
            ax.plot(self.k_knots, 1e2*median(df.iloc[:, self._tsa._sl_rratios].values, 0)**2, 'k.')
        setp(ax, ylabel='Transit depth [%]', xlabel=r'Wavelength [$\mu$m]', xlim=(self.data.wlmin, self.data.wlmax), ylim=ylim)

        if plot_resolution:
            yl = ax.get_ylim()
            ax.vlines(wavelength, yl[0], yl[0]+0.02*(yl[1]-yl[0]), ec='k')
        if xscale is not None:
            ax.set_xscale(xscale)
        if xticks is not None:
            ax.set_xticks(xticks, labels=xticks)
        return ax.get_figure()

    def plot_limb_darkening_parameters(self, result: Optional[str] = None, axs: Optional[tuple[Axes, Axes]] = None) -> None | Figure:
        """Plot the limb darkening parameters.

        Parameters
        ----------
        result
            The type of result to plot. Can be 'fit', 'mcmc', or None. If None, the default behavior is to use 'mcmc' if
            the MCMC sampler has been run, otherwise 'fit'. Default is None.
        axs
            The axes to plot the limb darkening parameters on. If None, a new figure with subplots will be created.
            Default is None.

        Returns
        -------
        Figure
            The figure containing the plot of the limb darkening parameters.

        Raises
        ------
        ValueError
            If the limb darkening model is not supported.
            If the result is not 'fit', 'mcmc', or None.
            If the result is 'mcmc' and the MCMC sampler has not been run.

        Notes
        -----
        This method plots the limb darkening parameters for two-parameter limb darkening models. It supports only
        quadratic, quadratic-tri, power-2, and power-2-pm models.
        """
        if not self._tsa.ldmodel in ('quadratic', 'quadratic-tri', 'power-2', 'power-2-pm'):
            return None

        if axs is None:
            fig, axs = subplots(1, 2, sharey='all', figsize=(13,4))
        else:
            fig = axs[0].get_figure()

        ldp1 = array([[p.prior.mean, p.prior.std] for p in self.ps[self._tsa._sl_ld][::2]])
        ldp2 = array([[p.prior.mean, p.prior.std] for p in self.ps[self._tsa._sl_ld][1::2]])

        if result is None:
            result = 'mcmc' if self._tsa.sampler is not None else 'fit'
        if result not in ('fit', 'mcmc'):
            raise ValueError("Result must be either 'fit', 'mcmc', or None")
        if result == 'mcmc' and self._tsa.sampler is None:
            raise ValueError("Cannot plot posterior solution before running the MCMC sampler.")

        wavelength = concatenate(self.wavelength)
        ix = argsort(wavelength)

        if result == 'fit':
            pv = self._tsa._de_population[self._tsa._de_imin]
            ldc = squeeze(concatenate(self._tsa._eval_ldc(pv), axis=1))
            axs[0].plot(self._tsa.ld_knots, pv[self._tsa._sl_ld][0::2], 'ok')
            axs[0].plot(wavelength[ix], ldc[:,0][ix])
            axs[1].plot(self._tsa.ld_knots, pv[self._tsa._sl_ld][1::2], 'ok')
            axs[1].plot(wavelength[ix], ldc[:,1][ix])
        else:
            pvp = self._tsa._mc_chains.reshape([-1, self._tsa.ndim])
            ldc = pvp[:,self._tsa._sl_ld]

            ld1m = median(ldc[:,::2], 0)
            ld1e = ldc[:,::2].std(0)
            ld2m = median(ldc[:,1::2], 0)
            ld2e = ldc[:,1::2].std(0)

            ldc = concatenate(self._tsa._eval_ldc(pvp), axis=1)
            ld1p = percentile(ldc[:,:,0], [50, 16, 84], axis=0)
            ld2p = percentile(ldc[:,:,1], [50, 16, 84], axis=0)

            axs[0].fill_between(wavelength[ix], ld1p[1, ix], ld1p[2, ix], alpha=0.5)
            axs[0].plot(wavelength[ix], ld1p[0][ix], 'k')
            axs[1].fill_between(wavelength[ix], ld2p[1, ix], ld2p[2, ix], alpha=0.5)
            axs[1].plot(wavelength[ix], ld2p[0][ix], 'k')

            axs[0].errorbar(self._tsa.ld_knots, ld1m, ld1e, fmt='ok')
            axs[1].errorbar(self._tsa.ld_knots, ld2m, ld2e, fmt='ok')

        axs[0].plot(self._tsa.ld_knots, ldp1[:,0] + ldp1[:,1], ':', c='C0')
        axs[0].plot(self._tsa.ld_knots, ldp1[:,0] - ldp1[:,1], ':', c='C0')
        axs[1].plot(self._tsa.ld_knots, ldp2[:,0] + ldp2[:,1], ':', c='C0')
        axs[1].plot(self._tsa.ld_knots, ldp2[:,0] - ldp2[:,1], ':', c='C0')

        setp(axs, xlim=(wavelength.min(), wavelength.max()), xlabel=r'Wavelength [$\mu$m]')
        setp(axs[0], ylabel='Limb darkening coefficient 1')
        setp(axs[1], ylabel='Limb darkening coefficient 2')
        return fig

    def plot_residuals(self, result: Optional[str] = None, ax: None | Axes | Sequence[Axes] = None,
                       pmin: float = 1, pmax: float = 99,
                       show_names: bool = False, cmap = None) -> Figure:
        """Plot the model residuals.

        Parameters
        ----------
        result
            The result type to plot. Must be either 'fit', 'mcmc', or None.
        ax
            The axes object to plot on. If None, a new figure and axes will be created.
        pmin
            The lower percentile to use when setting the color scale of the residuals image.
        pmax
            The upper percentile to use when setting the color scale of the residuals image.

        Returns
        -------
        Figure
            The figure object containing the plotted residuals.

        Raises
        ------
        ValueError
            If result is not one of 'fit', 'mcmc', or None.
        ValueError
            If result is 'mcmc' but the MCMC sampler has not been run.

        """
        if result not in ('fit', 'mcmc', None):
            raise ValueError("Result must be either 'fit', 'mcmc', or None")

        if result is None:
            if self._tsa._mc_chains is not None:
                result = 'mcmc'
            elif self._tsa._de_population is not None:
                result = 'fit'
            else:
                raise ValueError("Cannot plot residuals before running either the optimizer or the MCMC sampler.")

        if isinstance(self.data, TSData):
            nrows = 1
        else:
            nrows = self.data.ngroups

        if ax is None:
            fig, axs = subplots(nrows, 1, squeeze=False)
            axs = axs[:, 0]
        else:
            axs = [ax] if isinstance(ax, Axes) else ax
            if len(axs) != self.data.ngroups:
                raise ValueError("The number of axes must match the number of groups in the data.")
            fig = axs[0].figure

        if result == 'fit':
            pv = self._tsa._de_population[self._tsa._de_imin]
        else:
            pv = median(self._tsa._mc_chains.reshape([-1, self._tsa.ndim]), 0)

        fmodel = self._tsa.flux_model(pv)
        for ids, data in enumerate(self.data):
            ax = axs[ids]
            residuals = data.fluxes - squeeze(fmodel[ids])
            pp = percentile(residuals, [pmin, pmax])
            data.plot(ax=ax, data=residuals, vmin=pp[0], vmax=pp[1], cmap=cmap)

            tc = pv[1] + pv[2]*epoch(data.time.mean(), pv[1], pv[2])
            td = self.transit_duration

            #for i in range(2):
            #    ax.axvline(tc + (-1) ** i * 0.5 * td - self._tref, c='w', ymax=0.05, lw=5)
            #    ax.axvline(tc + (-1) ** i * 0.5 * td - self._tref, c='w', ymin=0.95, lw=5)
            #    ax.axvline(tc + (-1) ** i * 0.5 * td - self._tref, c='k', ymax=0.05, lw=1)
            #    ax.axvline(tc + (-1) ** i * 0.5 * td - self._tref, c='k', ymin=0.95, lw=1)
            if not show_names:
                ax.set_title("")

        if isinstance(fig, Figure):
            fig.tight_layout()
        return fig

    def plot_fit(self, result: Optional[str] = None, figsize: Optional[tuple[float, float]]=None, res_args=None, trs_args=None) -> Figure:
        """Plot either the best-fit model or the posterior model.

        Parameters
        ----------
        result
            Should be "fit", "mcmc", or None. Default is None.
        figsize
            The size of the figure in inches. Default is None.
        res_args
            Additional arguments for plotting residuals. Default is None.
        trs_args
            Additional arguments for plotting transmission spectrum. Default is None.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The plotted figure containing the residual plot, transmission spectrum plot, and limb darkening parameters plot.
        """
        if trs_args is None:
            trs_args = {}
        if res_args is None:
            res_args = {}

        fig = figure(layout='constrained', figsize=figsize)
        fts, fbelow = fig.subfigures(2, 1, hspace=0.07)
        fres, fldc = fbelow.subfigures(1, 2, wspace=0.05, width_ratios=(0.4, 0.6))
        axts = fts.add_subplot()
        axs_res = [fres.add_subplot(self.data.ngroups, 1, i+1) for i in range(self.data.ngroups)]
        axs_ldc = (fldc.add_subplot(2, 1, 1), fldc.add_subplot(2, 1, 2))

        self.plot_transmission_spectrum(result=result, ax=axts, **trs_args)
        self.plot_residuals(result=result, ax=axs_res, **res_args)
        self.plot_limb_darkening_parameters(result=result, axs=axs_ldc)
        fts.suptitle('Transmission spectrum')
        fres.suptitle('Residuals')
        fldc.suptitle('Limb darkening')
        setp(axs_ldc[0].get_xticklabels(), visible=False)
        for i in range(self.data.ngroups-1):
            setp(axs_res[i].get_xticklabels(), visible=False)
        setp(axs_ldc[0], xlabel="", ylabel='LDC 1', ylim=(0.18, 1.02))
        setp(axs_ldc[1], ylabel='LDC 2', ylim=(0.18, 1.02))
        fig.align_labels()
        return fig

    @property
    def transmission_spectrum(self) -> Table:
        """Get the posterior transmission spectrum as a Pandas DataFrame.

        Raises
        ------
        ValueError
            If the MCMC sampler has not been run before calculating the transmission spectrum.
        """
        if self._tsa._mc_chains is None:
            raise ValueError("Cannot calculate posterior transmission spectrum before running the MCMC sampler.")

        pvp = self.posterior_samples
        wls = concatenate(self.data.wavelength)
        ks = concatenate(self._tsa._eval_k(pvp.values[:, self._tsa._sl_rratios]), axis=1)
        ar = ks**2
        ix = argsort(wls)
        return Table(data=[wls[ix]*u.micrometer,
                           median(ks, 0)[ix], ks.std(0)[ix],
                           median(ar, 0)[ix], ar.std(0)[ix]],
                     names = ['wavelength', 'radius_ratio', 'radius_ratio_e', 'area_ratio', 'area_ratio_e'])

    def save(self, overwrite: bool = False) -> None:
        """Save the ExoIris analysis to a FITS file.

        Parameters
        ----------
        overwrite
            Flag indicating whether to overwrite an existing file with the same name.
        """
        pri = pf.PrimaryHDU()
        pri.header['name'] = self.name
        pri.header['t0'] = self.transit_center
        pri.header['t14'] = self.transit_duration
        pri.header['ndgroups'] = self.data.ngroups

        pr = pf.ImageHDU(name='priors')
        priors = [pickle.dumps(p) for p in self.ps]
        pr.header['priors'] = json.dumps(codecs.encode(pickle.dumps(priors), "base64").decode())

        if isinstance(self._tsa.ldmodel, LDTkLD):
            ldm = self._tsa.ldmodel
            pri.header['ldmodel'] = 'ldtk'
            pri.header['ldtkld'] = json.dumps(codecs.encode(pickle.dumps((ldm.sc.filters, ldm.sc.teff, ldm.sc.logg,
                                                                          ldm.sc.metal, ldm.dataset)), "base64").decode())
        else:
            pri.header['ldmodel'] = self._tsa.ldmodel

        k_knots = pf.ImageHDU(self._tsa.k_knots, name='k_knots')
        ld_knots = pf.ImageHDU(self._tsa.ld_knots, name='ld_knots')
        hdul = pf.HDUList([pri, k_knots, ld_knots, pr])

        for i, d in enumerate(self.data.data):
            flux = pf.ImageHDU(d.fluxes, name=f'flux_{i}')
            flux.header['name'] = d.name
            ferr = pf.ImageHDU(d.errors, name=f'ferr_{i}')
            wave = pf.ImageHDU(d.wavelength, name=f'wavelength_{i}')
            time = pf.ImageHDU(d.time, name=f'time_{i}')
            hdul.extend([time, wave, flux, ferr])

        if self._tsa.de is not None:
            de = pf.BinTableHDU(Table(self._tsa._de_population, names=self.ps.names), name='DE')
            de.header['npop'] = self._tsa.de.n_pop
            de.header['ndim'] = self._tsa.de.n_par
            de.header['imin'] = self._tsa.de.minimum_index
            hdul.append(de)

        if self._tsa.sampler is not None:
            mc = pf.BinTableHDU(Table(self._tsa.sampler.flatchain, names=self.ps.names), name='MCMC')
            mc.header['npop'] = self._tsa.sampler.nwalkers
            mc.header['ndim'] = self._tsa.sampler.ndim
            hdul.append(mc)

        hdul.writeto(f"{self.name}.fits", overwrite=True)

    def create_initial_population(self, n: int, source: str, add_noise: bool = True) -> ndarray:
        """Create an initial parameter vector population for the DE optimisation.

        Parameters
        ----------
        n
            Number of parameter vectors in the population.
        source
            Source of the initial population. Must be either 'fit' or 'mcmc'.
        add_noise
            Flag indicating whether to add noise to the initial population. Default is True.

        Returns
        -------
        ndarray
            The initial population.

        Raises
        ------
        ValueError
            If the source is not 'fit' or 'mcmc'.
        """
        return self._tsa.create_initial_population(n, source, add_noise)

    def add_noise_to_solution(self, result: str = 'fit') -> None:
        """Add noise to the global optimization result or MCMC parameter posteriors.

        Add noise to the global optimization result or MCMC parameter posteriors. You may want to do this if you
        create a new analysis from another one, for example, by adding radius ratio knots or changing the intrinsic
        data resolution.

        Parameters
        ----------
        result
            Determines which result to add noise to. Default is 'fit'.

        Raises
        ------
        ValueError
            If the 'result' argument is not 'fit' or 'mcmc'.
        """
        if result == 'fit':
            pvp = self._tsa._de_population[:, :].copy()
        elif result == 'mcmc':
            pvp = self._tsa._mc_chains[:, -1, :].copy()
        else:
            raise ValueError("The 'result' argument must be either 'fit' or 'mcmc'")

        npv = pvp.shape[0]
        pvp[:, 0] += normal(0, 0.005, size=npv)
        pvp[:, 1] += normal(0, 0.001, size=npv)
        pvp[:, 3] += normal(0, 0.005, size=npv)
        pvp[:, self._tsa._sl_rratios] += normal(0, 1, pvp[:, self._tsa._sl_rratios].shape) * 0.002 * pvp[:, self._tsa._sl_rratios]
        pvp[:, self._tsa._sl_ld] += normal(0, 1, pvp[:, self._tsa._sl_ld].shape) * 0.002 * pvp[:, self._tsa._sl_ld]

        if result == 'fit':
            self._tsa._de_population[:, :] = pvp
        else:
            pvp = self._tsa._mc_chains[:, -1, :] = pvp

    def optimize_gp_hyperparameters(self,
                                    log10_sigma_bounds: float | tuple[float, float] = (-5, -2),
                                    log10_rho_bounds: float | tuple[float, float] = (-5, 0),
                                    log10_sigma_prior=None, log10_rho_prior=None,
                                    npop: int = 10, niter: int = 100, subset = None):
        """Optimize the Matern-3/2 kernel Gaussian Process hyperparameters.

        Parameters
        ----------
        log10_sigma_bounds
            The bounds for the log10 of the sigma hyperparameter. If float is provided, the parameter will be
            fixed to the given value. Default is (-5, -2).
        log10_rho_bounds
            The bounds for the log10 of the rho hyperparameter. If float is provided, the parameter will be fixed
            to the given value. Default is (-5, 0).
        log10_sigma_prior
            The prior distribution for the sigma hyperparameter expressed as an object with a `logpdf` method
            or as an iterable containing the mean and standard deviation of the prior distribution. Default is None.
        log10_rho_prior
            The prior distribution for the rho hyperparameter expressed as an object with a `logpdf` method
            or as an iterable containing the mean and standard deviation of the prior distribution. Default is None.
        npop
            The population size for the differential evolution optimizer. Default is 10.
        niter
            The number of iterations for the differential evolution optimization process. Default is 100.
        subset
            The subset used for the optimization process. If `subset` is a float, a random subset of size
            `0.5 * self.npb` is used. If `subset` is an iterable, it must contain the indices of the subset.
            Default is None.

        Returns
        -------
        tuple[float, float]
            The optimized values for the log10 of the sigma and rho hyperparameters.
        float
            The fitness value.

        Raises
        ------
        ValueError
            If `subset` is not an iterable or a float.
        ValueError
            If `log10_sigma_prior` is not an object with a `logpdf` method or iterable.
        ValueError
            If `log10_rho_prior` is not an object with a `logpdf` method or iterable.

        Notes
        -----
        - The Gaussian Process is reconfigured with the optimal hyperparameters. Any previous kernels are overwritten.
        """

        if self._tsa.noise_model != 'fixed_gp':
            raise ValueError("The noise model must be set to 'fixed_gp' before the hyperparameter optimization.")

        sb = log10_sigma_bounds if isinstance(log10_sigma_bounds, Sequence) else [log10_sigma_bounds-1, log10_sigma_bounds+1]
        rb = log10_rho_bounds if isinstance(log10_rho_bounds, Sequence) else [log10_rho_bounds-1, log10_rho_bounds+1]
        bounds = array([sb, rb])

        data = self.data[0]
        if subset is not None:
            if isinstance(subset, float):
                ids = sort(permutation(data.nwl)[:int(subset*data.nwl)])
            elif isinstance(subset, Sequence):
                ids = array(subset, int)
            else:
                raise ValueError("subset must be either an iterable or a float.")
        else:
            ids = arange(data.nwl)

        class DummyPrior:
            def logpdf(self, x):
                return 0.0

        if log10_sigma_prior is not None:
            if isinstance(log10_sigma_prior, Sequence):
                sp = norm(*log10_sigma_prior)
            elif hasattr(log10_sigma_prior, 'logpdf'):
                sp = log10_sigma_prior
            else:
                raise ValueError('Bad sigma_prior')
        else:
            sp = DummyPrior()

        if log10_rho_prior is not None:
            if isinstance(log10_rho_prior, Sequence):
                rp = norm(*log10_rho_prior)
            elif hasattr(log10_rho_prior, 'logpdf'):
                rp = log10_rho_prior
            else:
                raise ValueError('Bad rho_prior')
        else:
            rp = DummyPrior()

        npb = ids.size
        time = (tile(data.time[newaxis, data.ootmask], (npb, 1)) + arange(npb)[:, newaxis]).ravel()
        flux = (data.fluxes[ids, :][:, data.ootmask]).ravel() - 1
        ferr = (data.errors[ids, :][:, data.ootmask]).ravel()
        gp = GaussianProcess(terms.Matern32Term(sigma=flux.std(), rho=0.1))

        def nll(log10x):
            x = 10**log10x
            if any(log10x < bounds[:,0]) or any(log10x > bounds[:,1]):
                return inf
            gp.kernel = terms.Matern32Term(sigma=x[0], rho=x[1])
            gp.compute(time, yerr=ferr, quiet=True)
            return -(gp.log_likelihood(flux) + sp.logpdf(log10x[0]) + rp.logpdf(log10x[1]))

        de = DiffEvol(nll, bounds, npop, min_ptp=0.2)
        if isinstance(log10_sigma_bounds, float):
            de.population[:, 0] = log10_sigma_bounds
        if isinstance(log10_rho_bounds, float):
            de.population[:, 1] = log10_rho_bounds

        de.optimize(niter)
        x = de.minimum_location
        self._tsa.set_gp_hyperparameters(10**x[0], 10**x[1])
        return 10**x, de._fitness.ptp()



#  ExoIris: fast, flexible, and easy exoplanet transmission spectroscopy in Python.
#  Copyright (C) 2024 Hannu Parviainen
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.
from copy import deepcopy
from typing import Optional

from ldtk import BoxcarFilter, LDPSetCreator
from numba import njit, prange
from numpy import zeros, log, pi, linspace, inf, atleast_2d, newaxis, clip, arctan2, ones, floor, sum, concatenate, \
    sort, ndarray, zeros_like, array, tile, arange, squeeze
from numpy.random import default_rng
from celerite2 import GaussianProcess as GP, terms

from pytransit import TSModel, RRModel, LDTkLD, TransitAnalysis
from pytransit.lpf.logposteriorfunction import LogPosteriorFunction

from pytransit.orbits import as_from_rhop, i_from_ba, fold, i_from_baew, d_from_pkaiews, epoch
from pytransit.param import ParameterSet, UniformPrior as UP, NormalPrior as NP, GParameter
from scipy.interpolate import splev, splrep

from .tsdata import TSDataSet

NM_WHITE = 0
NM_GP_FIXED = 1
NM_GP_FREE = 2

noise_models = dict(white=NM_WHITE, fixed_gp=NM_GP_FIXED, free_gp=NM_GP_FREE)

@njit(parallel=True, cache=False)
def lnlike_normal(o, m, e):
    if m.ndim == 2:
        return -sum(log(e)) - 0.5 * o.size * log(2. * pi) - 0.5 * sum((o - m) ** 2 / e ** 2)
    if m.ndim == 3:
        npv = m.shape[0]
        lnlike = zeros(npv)
        for i in prange(npv):
            lnlike[i] = -sum(log(e)) - 0.5 * o.size * log(2. * pi) - 0.5 * sum((o - m[i]) ** 2 / e ** 2)
        return lnlike


def resample(x_new, x_old, y_old):
    return splev(x_new, splrep(x_old, y_old, s=0.0))


def add_knots(x_new, x_old):
    return sort(concatenate([x_new, x_old]))


def clean_knots(knots, min_distance, lmin=0, lmax=inf):
    """Clean the knot table by replacing groups of adjacent knots with a single knot at the group mean.

    Parameters
    ----------
    knots : numpy.ndarray
        An array of knots.

    min_distance : float
        The minimum distance between adjacent knots.

    lmin : float, optional
        The minimum value of knots to consider. Default is 0.

    lmax : float, optional
        The maximum value of knots to consider. Default is inf.

    Returns
    -------
    numpy.ndarray
        An array of cleaned knots, where adjacent knots that are less than `min_distance` apart are replaced
        by the mean value of the group.
    """
    i = 0
    nknots = []
    while i < knots.size:
        m = [i]
        if lmin <= knots[i] <= lmax:
            j = i+1
            while i < knots.size - 1 and knots[j]-knots[i] < min_distance:
                j += 1
                i += 1
                m.append(i)
        nknots.append(knots[m].mean())
        i += 1
    return array(nknots)


class TSLPF(LogPosteriorFunction):
    def __init__(self, name: str, ldmodel, data: TSDataSet, nk: int = 50, nldc: int = 10, nthreads: int = 1,
                 tmpars = None, noise_model: str = 'white'):
        super().__init__(name)
        self._original_data: TSDataSet | None = None
        self.data: TSDataSet | None = None
        self.npb: list[int] | None= None
        self.npt: list[int] | None = None
        self.ndim: int | None = None

        self._gp: Optional[list[GP]] = None
        self._gp_time: Optional[list[ndarray]] = None
        self._gp_flux: Optional[list[ndarray]] = None

        self.set_noise_model(noise_model)

        self.ldmodel = ldmodel
        self.tms = [TSModel(ldmodel, nthreads=nthreads, **(tmpars or {})) for i in range(len(data))]
        self.set_data(data)

        if isinstance(ldmodel, LDTkLD):
            self.ldmodel._init_interpolation(self.tms[0].mu)

        self.nthreads = nthreads
        self.nldc = nldc
        self.nk = nk

        self.k_knots = linspace(data.wlmin, data.wlmax, self.nk)

        if isinstance(ldmodel, LDTkLD):
            self.ld_knots = array([])
        else:
            self.ld_knots = linspace(data.wlmin, data.wlmax, self.nldc)

        self._ootmask = None
        self._npv = 1
        self._de_population: Optional[ndarray] = None
        self._de_imin: Optional[int] = None
        self._mc_chains: Optional[ndarray] = None

        self._init_parameters()

    @property
    def flux(self) -> list[ndarray]:
        return self.data.fluxes

    @property
    def time(self) -> list[ndarray]:
        return self.data.time

    @property
    def wavelength(self) -> list[ndarray]:
        return self.data.wavelength

    @property
    def ferr(self) -> list[ndarray]:
        return self.data.errors

    def set_data(self, data: TSDataSet):
        self._original_data = deepcopy(data)
        self.data = data
        self.npb: list[int] = [f.shape[0] for f in self.flux]
        self.npt: list[int] = [f.shape[1] for f in self.flux]
        for i, time in enumerate(self.time):
            self.tms[i].set_data(time)
        if self._nm in (NM_GP_FIXED, NM_GP_FREE):
            self._init_gp()

    def _init_parameters(self) -> None:
        self.ps = ParameterSet([])
        self._init_p_star()
        self._init_p_orbit()
        self._init_limb_darkening()
        self._init_p_radius_ratios()
        self.ps.freeze()
        self.ndim = len(self.ps)

    def set_noise_model(self, noise_model: str) -> None:
        """Sets the noise model for the analysis.

        Parameters
        ----------
        noise_model : str
            The noise model to be used. Must be one of the following: white, fixed_gp, free_gp.

        Raises
        ------
        ValueError
            If noise_model is not one of the specified options.
        """
        if noise_model not in noise_models.keys():
            raise ValueError('noise_model must be one of: white, fixed_gp, free_gp')
        self.noise_model = noise_model
        self._nm = noise_models[noise_model]
        if self._nm in (NM_GP_FIXED, NM_GP_FREE):
            self._init_gp()

    def _init_gp(self) -> None:
        """Initializes the Gaussian Process (GP) .

        This method initializes the necessary variables and sets up the GP for the given data.
        """
        self._gp_time = []
        self._gp_flux = []
        self._gp_ferr = []
        self._gp = []
        for d in self.data:
            self._gp_time.append((tile(d.time[newaxis, :], (d.nwl, 1)) + arange(d.nwl)[:, newaxis]).ravel())
            self._gp_flux.append(d.fluxes.ravel())
            self._gp_ferr.append(d.errors.ravel())
            self._gp.append(GP(terms.Matern32Term(sigma=self._gp_flux[-1].std(), rho=0.1)))
            self._gp[-1].compute(self._gp_time[-1], yerr=self._gp_ferr[-1], quiet=True)

    def set_gp_hyperparameters(self, sigma: float, rho: float) -> None:
        """Sets the Gaussian Process hyperparameters assuming a Matern32 kernel.

        Parameters
        ----------
        sigma : float
            The kernel amplitude parameter.

        rho : float
            The length scale parameter.

        Raises
        ------
        RuntimeError
            If the GP has not been initialized before setting the hyperparameters.
        """
        if self._gp is None:
            raise RuntimeError('The GP needs to be initialized before setting hyperparameters.')
        for i, gp in enumerate(self._gp):
            gp.kernel = terms.Matern32Term(sigma=sigma, rho=rho)
            gp.compute(self._gp_time[i], yerr=self._gp_ferr[i], quiet=True)

    def set_gp_kernel(self, kernel: terms.Term) -> None:
        """Sets the kernel for the Gaussian Process (GP) model and recomputes the GP.

        Parameters
        ----------
        kernel : terms.Term
            The kernel to be set for the GP.
        """
        for i, gp in enumerate(self._gp):
            gp.kernel = kernel
            gp.compute(self._gp_time[i], yerr=self._gp_ferr[i], quiet=True)

    def _init_p_star(self) -> None:
        pstar = [GParameter('rho', 'stellar density', 'g/cm^3', UP(0.1, 25.0), (0, inf))]
        self.ps.add_global_block('star', pstar)

    def _init_limb_darkening(self) -> None:
        if isinstance(self.ldmodel, LDTkLD):
            pld = [GParameter('teff', 'stellar TEff', 'K', NP(*self.ldmodel.sc.teff), (0, inf)),
                   GParameter('logg', 'stellar log g', '', NP(*self.ldmodel.sc.logg), (0, inf)),
                   GParameter('metal', 'stellar metallicity', '', NP(*self.ldmodel.sc.metal), (-inf, inf))]
        else:
            pld = concatenate([
                [GParameter(f'ldc1_{l:08.5f}', fr'ldc1 at {l:08.5f} $\mu$m', '', UP(0, 1), bounds=(-inf, inf)),
                 GParameter(f'ldc2_{l:08.5f}', fr'ldc2 at {l:08.5f} $\mu$m', '', UP(0, 1), bounds=(-inf, inf))]
                for l in self.ld_knots])
        self.ps.add_global_block('limb_darkening', pld)
        self._start_ld = self.ps.blocks[-1].start
        self._sl_ld = self.ps.blocks[-1].slice

    def _init_p_orbit(self):
        ps = self.ps
        pp = [GParameter('tc', 'zero_epoch', '', NP(0.0, 0.1), (-inf, inf)),
              GParameter('p', 'period', 'd', NP(1.0, 1e-5), (0, inf)),
              GParameter('b', 'impact_parameter', 'R_s', UP(0.0, 1.0), (0, inf)),
              GParameter('secw', 'sqrt(e) cos(w)', '', NP(0.0, 1e-5), (-1, 1)),
              GParameter('sesw', 'sqrt(e) sin(w)', '', NP(0.0, 1e-5), (-1, 1))]
        ps.add_global_block('orbit', pp)
        self._start_orbit = ps.blocks[-1].start
        self._sl_orbit = ps.blocks[-1].slice

    def _init_p_radius_ratios(self):
        ps = self.ps
        pp = [GParameter(f'k_{k:08.5f}', fr'radius ratio at {k:08.5f} $\mu$m', 'A_s', UP(0.02, 0.2), (0, inf)) for k in self.k_knots]
        ps.add_global_block('radius_ratios', pp)
        self._start_rratios = ps.blocks[-1].start
        self._sl_rratios = ps.blocks[-1].slice

    def set_ldtk_prior(self, teff, logg, metal, dataset: str = 'visir-lowres', width: float = 50,
                       uncertainty_multiplier: float = 10):
        hw = 0.5 * width
        filters = [BoxcarFilter('a', wlc - hw, wlc + hw) for wlc in 1e3 * self.ld_knots]
        sc = LDPSetCreator(teff, logg, metal, filters=filters, dataset=dataset)
        ps = sc.create_profiles()

        match self.ldmodel:
            case 'power-2':
                ldc, lde = ps.coeffs_p2()
            case 'quadratic':
                ldc, lde = ps.coeffs_qd()
            case 'quadratic-triangular':
                ldc, lde = ps.coeffs_tq()
            case _:
                raise ValueError('Unsupported limb darkening model.')

        for i,l in enumerate(self.ld_knots):
            self.set_prior(f'ldc1_{l:08.5f}', 'NP', ldc[i, 0].round(3), (uncertainty_multiplier * lde[i, 0]).round(3))
            self.set_prior(f'ldc2_{l:08.5f}', 'NP', ldc[i, 1].round(3), (uncertainty_multiplier * lde[i, 1]).round(3))

    def add_k_knots(self, knot_wavelengths) -> None:
        """Add radius ratio (k) knots to the model.

        Parameters
        ----------
        knot_wavelengths : array-like
            An array of knot wavelengths to be added.
        """
        self.set_k_knots(concatenate([self.k_knots, knot_wavelengths]))

    def set_k_knots(self, knot_wavelengths) -> None:
        """Set the radius ratio (k) knot wavelengths for the model.

        Parameters
        ----------
        knot_wavelengths : array-like
            Array of knot wavelengths.

        """

        # Save the old variables
        # ----------------------
        xo = self.k_knots
        pso = self.ps
        deo = self._de_population
        mco = self._mc_chains
        slo = self._sl_rratios
        ndo = self.ndim

        xn = self.k_knots = clean_knots(sort(knot_wavelengths), 1e-5)
        self.nk = self.k_knots.size

        self._init_parameters()
        psn = self.ps
        sln = self._sl_rratios
        ndn = self.ndim

        # Set the priors back as they were
        # --------------------------------
        for po in pso:
            if po.name in psn.names:
                self.set_prior(po.name, po.prior)

        # Resample the DE parameter population
        # ------------------------------------
        if self._de_population is not None:
            den = zeros((deo.shape[0], ndn))

            # Copy the old parameter values
            # -----------------------------
            for pid_old, p in enumerate(pso):
                if p.name in psn.names:
                    pid_new = psn.find_pid(p.name)
                    den[:, pid_new] = deo[:, pid_old]

            # Resample the radius ratios
            # --------------------------
            for i in range(den.shape[0]):
                den[i, sln] = resample(xn, xo, deo[i, slo])

            self._de_population = den
            self.de = None

        # Resample the MCMC parameter population
        # --------------------------------------
        if self._mc_chains is not None:
            fmco = mco.reshape([-1, ndo])
            fmcn = zeros((fmco.shape[0], ndn))

            # Copy the old parameter values
            # -----------------------------
            for pid_old, p in enumerate(pso):
                if p.name in psn.names:
                    pid_new = psn.find_pid(p.name)
                    fmcn[:, pid_new] = fmco[:, pid_old]

            # Resample the radius ratios
            # --------------------------
            for i in range(fmcn.shape[0]):
                fmcn[i, sln] = resample(xn, xo, fmco[i, slo])

            self._mc_chains = fmcn.reshape([mco.shape[0], mco.shape[1], ndn])
            self.sampler = None


    def add_ld_knots(self, knot_wavelengths) -> None:
        """Add limb darkening knots to the model.

        Parameters
        ----------
        knot_wavelengths : array-like
            An array of knot wavelengths to be added.
        """
        self.set_ld_knots(concatenate([self.ld_knots, knot_wavelengths]))

    def set_ld_knots(self, knot_wavelengths) -> None:
        """Set the limb darkening knot wavelengths for the model.

        Parameters
        ----------
        knot_wavelengths : array-like
            Array of knot wavelengths.
        """
        xo = self.ld_knots
        xn = self.ld_knots = sort(knot_wavelengths)
        self.nldc = self.ld_knots.size

        pvpo = self.de.population.copy() if self.de is not None else None
        pso = self.ps
        sldo = self._sl_ld
        self._init_parameters()
        psn = self.ps
        sldn = self._sl_ld
        for po in pso:
            if po.name in psn.names:
                self.set_prior(po.name, po.prior)

        if self.de is not None:
            pvpn = self.create_pv_population(pvpo.shape[0])
            # Copy the old parameter values
            # -----------------------------
            for pid_old, p in enumerate(pso):
                if p.name in psn:
                    pid_new = psn.find_pid(p.name)
                    pvpn[:, pid_new] = pvpo[:, pid_old]

            # Resample the radius ratios
            # --------------------------
            for i in range(pvpn.shape[0]):
                pvpn[i, sldn] = resample(xn, xo, pvpo[i, sldo])

            self.de = None
            self._de_population = pvpn

    def _eval_k(self, pvp):
        """
        Evaluate the radius ratio model.

        Parameters
        ----------
        pvp : ndarray
            The input array of shape (npv, np), where npv is the number of parameter vectors
            and np is the number of parameters.

        Returns
        -------
        ks : ndarray
            The radius ratios of shape (npv, npb), where npb is the number of passbands.
        """
        pvp = atleast_2d(pvp)
        ks = [zeros((pvp.shape[0], npb)) for npb in self.npb]
        for ids in range(self.data.ngroups):
            for ipv in range(pvp.shape[0]):
                ks[ids][ipv,:] =  splev(self.wavelength[ids], splrep(self.k_knots, pvp[ipv], s=0.0))
        return ks

    def _eval_ldc(self, pvp):
        if isinstance(self.ldmodel, LDTkLD):
            ldp = pvp[:, newaxis, self._sl_ld]
            ldp[:, 0, 0] = clip(ldp[:, 0, 0], *self.ldmodel.sc.client.teffl)
            ldp[:, 0, 1] = clip(ldp[:, 0, 1], *self.ldmodel.sc.client.loggl)
            ldp[:, 0, 2] = clip(ldp[:, 0, 2], *self.ldmodel.sc.client.zl)
            return ldp
        else:
            pvp = atleast_2d(pvp)
            ldk = pvp[:, self._sl_ld].reshape([pvp.shape[0], self.nldc, 2])
            ldp = [zeros((pvp.shape[0], npb, 2)) for npb in self.npb]
            for ids in range(self.data.ngroups):
                for ipv in range(pvp.shape[0]):
                    ldp[ids][ipv, :, 0] = splev(self.wavelength[ids], splrep(self.ld_knots, ldk[ipv, :, 0], s=0.0))
                    ldp[ids][ipv, :, 1] = splev(self.wavelength[ids], splrep(self.ld_knots, ldk[ipv, :, 1], s=0.0))
            return ldp

    def transit_model(self, pv, copy=True):
        """Evaluates the transit model for parameter vector pv.

        Parameters
        ----------
        pv : numpy.ndarray
            Array of transit parameters. Each row represents a set of transit parameters for a single transit event.
            The columns of the array should be in the following order:
            - Column 0: stellar density (g/cm^3)
            - Column 1: transit center time (T0)
            - Column 2: orbital period (P)
            - Column 3: impact parameter
            - Column 4: sqrt e cos w
            - Column 5: sqrt e sin w
            - Column 6: planet-to-star radius ratio (Rp/R_star)

        copy : bool, optional
            Whether to create a copy of the calculated values before returning the result. Default is True.

        Returns
        -------
        numpy.ndarray
            Array of model transit fluxes. Each element corresponds to the transit model flux for the corresponding set
            of transit parameters in the input array 'pv'.
        """
        pv = atleast_2d(pv)
        ldp = self._eval_ldc(pv)
        t0 = pv[:, 1]
        p = pv[:, 2]
        k = self._eval_k(pv[:, self._sl_rratios])
        aor = as_from_rhop(pv[:, 0], p)
        inc = i_from_ba(pv[:, 3], aor)
        ecc = pv[:, 4] ** 2 + pv[:, 5] ** 2
        w = arctan2(pv[:, 5], pv[:, 4])
        return [tm.evaluate(k[i], ldp[i], t0, p, aor, inc, ecc, w, copy) for i,tm in enumerate(self.tms)]

    def flux_model(self, pv):
        return self.transit_model(pv)

    def create_pv_population(self, npop: int = 50):
        """ Crate a parameter vector population.
        Parameters
        ----------
        npop : int, optional
            The number of parameter vectors in the population. Default is 50.

        Returns
        -------
        population : array_like
            An array of parameter vectors sampled from the prior distribution.
        """
        return self.ps.sample_from_prior(npop)

    def set_radius_ratio_limits(self, kmin, kmax):
        for ipb in range(self.nk):
            self.set_prior(f'k_{ipb + 1:03d}', 'UP', kmin, kmax)

    def lnlikelihood(self, pv):
        """Log likelihood for parameter vector pv.

        Parameters
        ----------
        pv : array-like
            The input parameter values for the flux model.

        Returns
        -------
        lnlike : float
            The logarithm of the likelihood value calculated using the normal distribution.

        """
        pv = atleast_2d(pv)
        npv = pv.shape[0]
        fmod = self.flux_model(pv)
        lnl = zeros(npv)
        if self._nm == NM_WHITE:
            for i in range(self.data.ngroups):
                lnl += lnlike_normal(self.flux[i], fmod[i], self.ferr[i])
        elif self._nm == NM_GP_FIXED:
            for j in range(npv):
                for i in range(self.data.ngroups):
                    lnl[j] += self._gp[i].log_likelihood(self._gp_flux[i] - fmod[i][j].ravel())
        else:
            raise NotImplementedError("The free GP noise model hasn't been implemented yet.")
        return lnl if npv > 1 else lnl[0]

    def create_initial_population(self, n: int, source: str, add_noise: bool = True) -> ndarray:
        """Create an initial parameter vector population for DE.

        Parameters
        ----------
        n : int
            Number of parameter vectors in the population.
        source : str
            Source of the initial population. Must be either 'fit' or 'mcmc'.
        add_noise : bool, optional
            Flag indicating whether to add noise to the initial population. Default is True.

        Returns
        -------
        numpy.ndarray
            The initial population.

        Raises
        ------
        ValueError
            If the source is not 'fit' or 'mcmc'.
        """
        rng = default_rng()

        if source not in ('fit', 'mcmc'):
            raise ValueError("source must be either 'fit' or 'mcmc'")

        if source == 'fit':
            pvs = self._de_population
            if n == pvs.shape[0]:
                pvp = pvs.copy()
            else:
                pvp = rng.choice(pvs, size=n)
        else:
            pvs = self._mc_chains
            if n == pvs.shape[2]:
                pvp = pvs[:,-1,:].copy()
            else:
                pvp = rng.choice(pvs.reshape([-1, self.ndim]), size=n)

        if add_noise:
            pvp[:, 0] += rng.normal(0, 0.005, size=n)
            pvp[:, 1] += rng.normal(0, 0.001, size=n)
            pvp[:, 3] += rng.normal(0, 0.005, size=n)
            pvp[:, self._sl_rratios] += rng.normal(0, 1, pvp[:, self._sl_rratios].shape) * 0.002 * pvp[:, self._sl_rratios]
            pvp[:, self._sl_ld] += rng.normal(0, 1, pvp[:, self._sl_ld].shape) * 0.002 * pvp[:, self._sl_ld]
        return pvp

    def optimize_global(self, niter=200, npop=50, population=None, pool=None, lnpost=None, vectorize=True,
                        label='Global optimisation', leave=False, plot_convergence: bool = True, use_tqdm: bool = True,
                        plot_parameters: tuple = (0, 2, 3, 4), min_ptp: float = 5):

        if population is None:
            if self._de_population is None:
                population = self.create_pv_population()
            else:
                population = self._de_population

        super().optimize_global(niter=niter, npop=npop, population=population, pool=pool, lnpost=lnpost,
                                vectorize=vectorize, label=label, leave=leave, plot_convergence=plot_convergence,
                                use_tqdm=use_tqdm, plot_parameters=plot_parameters, min_ptp=min_ptp)
        self._de_population = self.de.population.copy()
        self._de_imin = self.de.minimum_index

    def sample_mcmc(self, niter: int = 500, thin: int = 5, repeats: int = 1, npop: int = None, population=None,
                    label='MCMC sampling', reset=True, leave=True, save=False, use_tqdm: bool = True, pool=None,
                    lnpost=None, vectorize: bool = True):

        if population is None:
            if self._mc_chains is None:
                population = self._de_population.copy()
            else:
                population = self._mc_chains[:, -1, :].copy()

        super().sample_mcmc(niter, thin, repeats, npop=npop, population=population, label=label, reset=reset,
                            leave=leave, save=save, use_tqdm=use_tqdm, pool=pool, lnpost=lnpost, vectorize=vectorize)
        self._mc_chains = self.sampler.chain.copy()



import pytest
from exoiris.binning import Binning, CompoundBinning


def test_binning_nb():
    binning = Binning(0.5, 2.5, nb=4)
    assert binning.xmin == 0.5
    assert binning.xmax == 2.5
    assert binning.bw == 0.5
    assert binning.nb == 4
    assert binning.bins.shape == (4,2)


def test_binning_bw():
    binning = Binning(0.5, 2.5, bw=0.5)
    assert binning.xmin == 0.5
    assert binning.xmax == 2.5
    assert binning.bw == 0.5
    assert binning.nb == 4
    assert binning.bins.shape == (4,2)


def test_binning_r():
    binning = Binning(0.5, 2.5, r=10.0)
    assert binning.xmin == 0.5
    assert binning.xmax == 2.5
    assert binning.r == 10.0
    assert binning.bins.shape == (17,2)


def test_add_normal_binning():
    binning_1 = Binning(xmin=0.5, xmax=1.0, nb=5)
    binning_2 = Binning(xmin=1.0, xmax=2.0, nb=5)
    result = binning_1 + binning_2
    assert isinstance(result, CompoundBinning)


def test_add_compound_binning():
    binning = Binning(xmin=0.5, xmax=1.0, nb=5)
    compound = CompoundBinning([Binning(xmin=1.0, xmax=2.0, nb=5)])
    result = binning + compound
    assert isinstance(result, CompoundBinning)


def test_add_incompatible_type():
    binning = Binning(xmin=1.0, xmax=2.0, nb=5)
    with pytest.raises(TypeError):
        result = binning + "invalid type"





#  ExoIris: fast, flexible, and easy exoplanet transmission spectroscopy in Python.
#  Copyright (C) 2024 Hannu Parviainen
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numba import njit
from numpy import zeros, sum, sqrt, linspace, vstack, concatenate, floor, dot, ndarray, nan


@njit
def bin2d(v, e, el, er, bins, estimate_errors: bool = False) -> tuple[ndarray, ndarray]:
    """Bin 2D exoplanet transmission spectrum data with its uncertainties into predefined bins in wavelength.

        Parameters
        ----------
        v : ndarray
            A 2D array of the exoplanet transmission spectrum data with a shape (n_wavelength, n_exposure).
        e : ndarray
            A 2D array of uncertainties associated with the spectrum data in `v`, matching the shape of `v`.
        el : ndarray
            A 1D array containing the left wavelength edges of the integration ranges for each spectral data point.
        er : ndarray
            A 1D array containing the right wavelength edges of the integration ranges for each spectral data point.
        bins : ndarray
            A 2D array containing the edges of the wavelength bins. These should be sorted in ascending order.
        estimate_errors: bool, optional.
            Should the uncertainties be estimated from the data. Default value is False.

        Returns
        -------
        tuple of ndarrays
            A tuple containing two 2D arrays:
            - The first array (`bv`) contains the binned values of the transmission spectrum.
            - The second array (`be`) contains the binned uncertainties.
    """
    nbins = len(bins)
    ndata = v.shape[0]
    bv = zeros((nbins, v.shape[1]))
    be = zeros((nbins, v.shape[1]))
    e2 = e**2
    weights = zeros(ndata)
    i = 0
    for ibin in range(nbins):
        weights[:] = 0.0
        npt = 0
        bel, ber = bins[ibin]
        for i in range(i, ndata - 1):
            if el[i + 1] > bel:
                break
        il = i
        if er[i] > ber:
            weights[i] = ber - bel
            npt += 1
        else:
            weights[i] = er[i] - bel
            npt += 1
            for i in range(i + 1, ndata):
                if er[i] < ber:
                    weights[i] = er[i] - el[i]
                    npt += 1
                else:
                    weights[i] = ber - el[i]
                    npt += 1
                    break
        ir = i
        ws = sum(weights)
        bv[ibin] = vmean = dot(weights[il:ir+1], v[il:ir+1,:]) / ws
        if estimate_errors:
            if npt > 1:
                be[ibin] = sqrt(dot(weights[il:ir+1], (v[il:ir+1,:] - vmean)**2) / ws) / sqrt(npt)
            else:
                be[ibin] = nan
        else:
            be[ibin] = sqrt(dot(weights[il:ir+1], e2[il:ir+1,:])) / ws
    return bv, be


def create_binning(ranges, bwidths):
    """
    Create a combined array of discretization edges for multiple ranges with specified bin widths.

    This function generates a single concatenated array of discretization bins for given ranges
    where each range can have a different bin width. Each range is divided into bins of the specified width,
    and the start and end points of these bins are stored.

    Parameters
    ----------
    ranges : list of tuples
        A list where each tuple contains two elements (start, end) defining the range over which to create bins.
    bwidths : list of float
        A list of bin widths corresponding to each range in `ranges`. Each width specifies the size of the bin
        for the corresponding range.

    Returns
    -------
    ndarray
        A 2D NumPy array where each row contains the start and end points of a bin.

    Notes
    -----
    The final bin widths may differ from the given ones if they do not perfectly divide the ranges.
    """
    bins = []
    for r, w in zip(ranges, bwidths):
        n = int(floor((r[1] - r[0]) / w))
        e = linspace(*r, num=n)
        bins.append(vstack([e[:-1], e[1:]]).T)
    return concatenate(bins)


#  ExoIris: fast, flexible, and easy exoplanet transmission spectroscopy in Python.
#  Copyright (C) 2024 Hannu Parviainen
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import warnings
import numba

from collections.abc import Sequence

from typing import Union, Optional

from astropy.stats import mad_std
from matplotlib.figure import Figure
from matplotlib.pyplot import subplots, setp
from matplotlib.ticker import LinearLocator
from numpy import isfinite, median, where, concatenate, all, zeros_like, diff, asarray, interp, arange, floor, ndarray
from pytransit.orbits import fold
from scipy.ndimage import median_filter

from .util import bin2d
from .binning import Binning, CompoundBinning

class TSData:
    """
    TSData is a utility class representing transmission spectroscopy time series data with associated wavelength,
    fluxes, and errors. It provides methods for manipulating and analyzing the data.

    Attributes
    ----------
    time: 1D ndarray
        Array of time values.
    wavelength : 1D ndarray
        Array of wavelength values.
    fluxes : 2D ndarray
        2D array of flux values with a shape ``(nwl, npt)``, where ``nwl`` is the number of wavelengths and ``npt`` the
        number of exposures.
    errors : 2D ndarray
        2D Array of error values with a shape ``(nwl, npt)``, where ``nwl`` is the number of wavelengths and ``npt`` the
        number of exposures.
    """
    def __init__(self, time: Sequence, wavelength: Sequence, fluxes: Sequence, errors: Sequence,
                 name: str = "", wl_edges : Sequence | None = None, tm_edges : Sequence | None = None):
        """
        Parameters
        ----------
        time
            Array of time values.
        wavelength
            Array of wavelength values.
        fluxes
            2D array of flux values with a shape ``(nwl, npt)``, where ``nwl`` is the number of wavelengths and ``npt`` the
            number of exposures.
        errors
            2D Array of error values with a shape ``(nwl, npt)``, where ``nwl`` is the number of wavelengths and ``npt`` the
            number of exposures.
        wl_edges
            Tuple containing left and right wavelength edges for each wavelength element.
        """
        time, wavelength, fluxes, errors = asarray(time), asarray(wavelength), asarray(fluxes), asarray(errors)

        if fluxes.shape[0] != wavelength.size:
            raise ValueError("The size of the flux array's first axis must match the size of the wavelength array.")

        m = all(isfinite(fluxes), axis=1)
        self.name = name
        self.time = time.copy()
        self.wavelength = wavelength[m]
        self.fluxes = fluxes[m]
        self.errors = errors[m]
        self.ootmask = None
        self._update()
        self.groups = [slice(0, self.nwl)]

        if wl_edges is None:
            dwl = zeros_like(self.wavelength)
            dwl[:-1] = diff(self.wavelength)
            dwl[-1] = dwl[-2]
            self._wl_l_edges = self.wavelength - 0.5 * dwl
            self._wl_r_edges = self.wavelength + 0.5 * dwl
        else:
            self._wl_l_edges = wl_edges[0]
            self._wl_r_edges = wl_edges[1]

        if tm_edges is None:
            dt = zeros_like(self.time)
            dt[:-1] = diff(self.time)
            dt[-1] = dt[-2]
            self._tm_l_edges = self.time - 0.5 * dt
            self._tm_r_edges = self.time + 0.5 * dt
        else:
            self._tm_l_edges = tm_edges[0]
            self._tm_r_edges = tm_edges[1]

    def __repr__(self) -> str:
        return f"TSData Name:'{self.name}' [{self.wavelength[0]:.2f} - {self.wavelength[-1]:.2f}] nwl={self.nwl} npt={self.npt}"

    def calculate_ootmask(self, tc: float, p: float, t14: float):
        phase = fold(self.time, p, tc)
        self.ootmask = abs(phase) > 0.502 * t14

    def _update(self) -> None:
        """Update the internal attributes."""
        self.nwl = self.wavelength.size
        self.npt = self.time.size
        self.wllims = self.wavelength.min(), self.wavelength.max()

    def crop_wavelength(self, lmin: float, lmax: float) -> None:
        """Crop the data to include only the wavelength range between lmin and lmax.

        Parameters
        ----------
        lmin
            The minimum wavelength value to crop.
        lmax
            The maximum wavelength value to crop.

        Note
        ----
        The data will be modified in place.
        """
        m = (self.wavelength > lmin) & (self.wavelength < lmax)
        self.wavelength = self.wavelength[m]
        self.fluxes = self.fluxes[m]
        self.errors = self.errors[m]
        self._wl_l_edges = self._wl_l_edges[m]
        self._wl_r_edges = self._wl_r_edges[m]
        self._update()

    def remove_outliers(self, sigma: float = 5.0):
        """Remove outliers along the wavelength axis.

        Replace outliers along the wavelength axis with the value of a 5-point running median filter. Outliers are
        defined as data points that deviate from the median by more than sigma times the median absolute deviation
        along the wavelength axis.

        Parameters
        ----------
        sigma
            The number of standard deviations to use as the threshold for outliers.

        Note
        ----
        The data will be modified in place.
        """
        fm = median(self.fluxes, axis=0)
        fe = mad_std(self.fluxes, axis=0)
        self.fluxes = where(abs(self.fluxes - fm) / fe < sigma, self.fluxes, median_filter(self.fluxes, 5))

    def plot(self, ax=None, vmin: float = None, vmax: float = None, cmap=None, figsize=None, data=None) -> Figure:
        """Plot the spectroscopic light curves as a 2D image.

        Plot the spectroscopic light curves as a 2D image with time on the x-axis, wavelength and light curve index
        on the y-axis, and the flux as a color.

        Parameters
        ----------
        ax
            The subplot axes on which to plot. If None, a new figure and axes will be created.

        vmin
            The minimum value of the color scale.

        vmax
            The maximum value of the color scale.

        cmap
            The colormap to be used.

        figsize
            The size of the figure in inches (width, height).

        data
            Dataset to plot instead of self.fluxes.

        Returns
        -------
        Figure

        """
        if ax is None:
            fig, ax = subplots(figsize=figsize)
        else:
            fig = ax.figure
        tref = floor(self.time.min())

        def forward(x):
            return interp(x, self.wavelength, arange(self.nwl))
        def inverse(x):
            return interp(x, arange(self.nwl), self.wavelength)

        data = data if data is not None else self.fluxes
        ax.pcolormesh(self.time - tref, self.wavelength, data, vmin=vmin, vmax=vmax, cmap=cmap)
        setp(ax, ylabel=r'Wavelength [$\mu$m]', xlabel=f'Time - {tref:.0f} [BJD]')
        ax.yaxis.set_major_locator(LinearLocator(10))
        ax.yaxis.set_major_formatter('{x:.2f}')

        if self.name != "":
            ax.set_title(self.name)

        ax2 = ax.secondary_yaxis('right', functions=(forward, inverse))
        ax2.set_ylabel('Light curve index')
        ax2.set_yticks(forward(ax.get_yticks()))
        ax2.yaxis.set_major_formatter('{x:.0f}')
        return fig

    def __add__(self, other: 'TSData') -> 'TSDataSet':
        """Combine two transmission spectra along the wavelength axis.

        Parameters
        ----------
        other
            The TSData object to be added to the current TSData object.

        Returns
        -------
        TSDataSet
            The resulting TSDataSet object combining the two TSData objects.

        """
        return TSDataSet([self, other])

    def bin_wavelength(self, binning: Optional[Union[Binning, CompoundBinning]] = None,
                       nb: Optional[int] = None, bw: Optional[float] = None, r: Optional[float] = None,
                       estimate_errors: bool = False):
        """Bin the data along the wavelength axis.

        Bin the data along the wavelength axis. If binning is not specified, a Binning object is created using the
        minimum and maximum values of the wavelength.

        Parameters
        ----------
        binning
            The binning method to use. Default value is None.
        nb
            Number of bins. Default value is None.
        bw
            Bin width. Default value is None.
        r
            Bin resolution. Default value is None.
        estimate_errors
            Should the uncertainties be estimated from the data. Default value is False.

        Returns
        -------
        TSData
            The binned data.
        """
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', numba.NumbaPerformanceWarning)
            if binning is None:
                binning = Binning(self.wllims[0], self.wllims[1], nb=nb, bw=bw, r=r)
            bf, be = bin2d(self.fluxes, self.errors, self._wl_l_edges, self._wl_r_edges,
                           binning.bins, estimate_errors=estimate_errors)
            return TSData(self.time, binning.bins.mean(1), bf, be, wl_edges=(binning.bins[:,0], binning.bins[:,1]),
                          name=self.name, tm_edges=(self._tm_l_edges, self._tm_r_edges))


    def bin_time(self, binning: Optional[Union[Binning, CompoundBinning]] = None,
                       nb: Optional[int] = None, bw: Optional[float] = None, r: Optional[float] = None,
                       estimate_errors: bool = False):
        """Bin the data along the time axis.

        Bin the data along the time axis. If binning is not specified, a Binning object is created using the
        minimum and maximum time values.

        Parameters
        ----------
        binning
            The binning method to use. Default value is None.
        nb
            Number of bins. Default value is None.
        bw
            Bin width. Default value is None.
        r
            Bin resolution. Default value is None.
        estimate_errors
            Should the uncertainties be estimated from the data. Default value is False.

        Returns
        -------
        TSData
            The binned data.
        """
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', numba.NumbaPerformanceWarning)
            if binning is None:
                binning = Binning(self.time.min(), self.time.max(), nb=nb, bw=bw, r=r)
            bf, be = bin2d(self.fluxes.T, self.errors.T, self._tm_l_edges, self._tm_r_edges,
                           binning.bins, estimate_errors=estimate_errors)
            return TSData(binning.bins.mean(1), self.wavelength, bf.T, be.T,
                          wl_edges=(self._wl_l_edges, self._wl_r_edges),
                          tm_edges=(binning.bins[:,0], binning.bins[:,1]),
                          name=self.name)


class TSDataSet:
    """A high-level data storage class that can contain multiple TSData objects."""
    def __init__(self, data: Sequence[TSData]):
        self.data: list[TSData] = list(data)
        self.time: list[ndarray] = [d.time for d in data]
        self.fluxes: list[ndarray] = [d.fluxes for d in data]
        self.errors: list[ndarray] = [d.errors for d in data]
        self.wavelength: list[ndarray] = [d.wavelength for d in data]
        self.wlmin = concatenate(self.wavelength).min()
        self.wlmax = concatenate(self.wavelength).max()
        self.ngroups: int = len(self.data)
        self.groups: list = []
        i = 0
        for d in data:
            self.groups.append(slice(i, i+d.nwl))
            i += d.nwl

    def calculate_ootmask(self, tc: float, p: float, t14: float):
        for d in self.data:
            d.calculate_ootmask(tc, p, t14)

    def __getitem__(self, index: int) -> TSData:
        return self.data[index]

    def __len__(self) -> int:
        return self.ngroups

    def __repr__(self):
        return f"TSDataSet with {self.ngroups} groups: {str(self.groups)}"

    def plot(self, ax=None, vmin: float = None, vmax: float = None, cmap=None, figsize=None, data: ndarray | None = None):
        """Plot all the data sets.

        Parameters
        ----------
        ax
            The Axes object used for plotting. If None, a new set of subplots will be created.
        vmin
            The minimum value for the color mapping. Default is None.
        vmax
            The maximum value for the color mapping. Default is None.
        cmap
            The colormap used for mapping the data values to colors. Default is None.
        figsize
            The size of the figure created if `ax` is None. Default is None.
        data
            The data to be plotted. If None, the `self.data` attribute will be used.

        Returns
        -------
        Figure
            The Figure object that contains the subplots.

        """
        if ax is None:
            fig, ax = subplots(self.ngroups, 1, figsize=figsize, sharex='all')
        else:
            fig = ax[0].get_figure()

        if data is None:
            for i, a in enumerate(ax):
                self.data[i].plot(ax=a, vmin=vmin, vmax=vmax, cmap=cmap)
        else:
            for i, a in enumerate(ax):
                self.data[i].plot(ax=a, vmin=vmin, vmax=vmax, cmap=cmap, data=data[self.groups[i], :])

        setp(ax[:-1], xlabel='')
        return fig

    def __add__(self, other):
        if isinstance(other, TSData):
            return TSDataSet(self.data + [other])
        elif isinstance(other, TSDataSet):
            return TSDataSet(self.data + other.data)


#  ExoIris: fast, flexible, and easy exoplanet transmission spectroscopy in Python.
#  Copyright (C) 2024 Hannu Parviainen
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections.abc import Sequence

from numpy import array, floor, linspace, vstack, nan, concatenate, ndarray


class Binning:
    """Class representing a homogeneous binning within a given range.

    The `Binning` is defined by its minimum and maximum values (`xmin` and `xmax`), and either the number
    of bins (`nb`), bin width (`bw`), or resolving power (`r`). If the `Binning` is
    initialized giving the number of bins, the range will be divided into `nb` equally
    wide bins. If the `Binning` is initialized giving the bin width, the range will be
    divided into bins that are as close as `bw` as possible. Finally, if the `Binning` is
    initialized with the resolving power, the bin widths aim to follow the x/xv relation.

    Parameters
    ----------
    xmin
        The minimum value of the range of values to be binned.
    xmax
        The maximum value of the range of values to be binned.
    nb
        The number of bins to be used for binning the range of values.
    bw
        The bin width to be used for binning the range of values.
    r
        The resolving power  to be used for binning the range of values.

    Raises
    ------
    ValueError
        When none or more than one of `nb`, `bw`, and `r` are provided.

    Attributes
    ----------
    xmin
        The minimum value of the range of values to be binned.
    xmax
        The maximum value of the range of values to be binned.
    nb
        The number of bins.
    bw
        The bin width.
    r
        The resolving power.
    bins
        An array of left and right bin edge values.

    """

    def __init__(self, xmin: float, xmax: float,
                 nb: float | None = None,
                 bw: float | None = None,
                 r: float | None = None) -> None:

        if (nb is not None) + (bw is not None) + (r is not None) != 1:
            raise ValueError(
                'A binning needs to be initialized either with the number of bins (nb), bin width (bw), or resolution (r)')

        self.xmin: float = xmin
        self.xmax: float = xmax
        self.nb: int | None = nb
        self.bw: float | None = bw
        self.r: float | None = r

        self.bins : ndarray | None = None
        """A (nb, 2) array of left and right bin edge values."""

        if r is not None:
            self._bin_r()
        elif bw is not None:
            self._bin_bw()
        else:
            self._bin_nb()

    def _bin_r(self) -> None:
        """Create bins for a given range using the resolution."""
        bins = []
        x0 = self.xmin
        while True:
            x1 = x0 * (2 * self.r + 1) / (2 * self.r - 1)
            bins.append([x0, min(x1, self.xmax)])
            x0 = x1
            if x1 >= self.xmax:
                break
        self.bins = array(bins)
        self.nb = self.bins.shape[0]

    def _bin_bw(self) -> None:
        """Create bins for a given range based on the bin width `bw` and the range limits (xmin, xmax).

        Raises
        ------
        ValueError
            If the bin width (bw) is greater than or equal to the binning span.
        """
        if self.xmax - self.xmin <= self.bw:
            raise ValueError("The bin width (bw) should be smaller than the binning span.")
        self.nb = int(floor((self.xmax - self.xmin) / self.bw))
        self._bin_nb()
        self.nb = self.bins.shape[0]

    def _bin_nb(self) -> None:
        """Create bins for a given range based on the number of bins `nb` and the range limits (xmin, xmax).

        Raises
        ------
        ValueError
            If the number of bins (nb) is less than or equal to zero.
        """
        if self.nb <= 0:
            raise ValueError("The number of bins (nb) should be larger than zero.")
        e = linspace(self.xmin, self.xmax, num=self.nb+1)
        self.bins = vstack([e[:-1], e[1:]]).T
        self.bw = (self.xmax - self.xmin) / self.nb

    def __repr__(self) -> str:
        return f"Binning {self.xmin:0.4f} - {self.xmax:.4f}: dx = {self.bw or nan:6.4f}, n = {self.nb:4d}, R = {self.r}"

    def __add__(self, other: 'Binning') -> 'CompoundBinning':
        if isinstance(other, Binning):
            return CompoundBinning([self, other])
        elif isinstance(other, CompoundBinning):
            return other + self
        else:
            raise TypeError(f"Can't concatenate Binning and {other.__class__.__name__}")


class CompoundBinning:
    """A class representing complex heterogeneous binning.

    This class allows for creating a compound binning by combining multiple binning objects.

    Parameters
    ----------
    binnings
        The list of Binning objects used for creating the bins.

    Attributes
    ----------
    binnings
        List of binning objects.
    bins
        Array of bin edges from all binning objects.
    """
    def __init__(self, binnings: Sequence[Binning]) -> None:
        self.binnings = binnings
        self.bins = concatenate([b.bins for b in self.binnings])

    def __repr__(self):
        return 'CompoundBinning:\n' + '\n'.join([b.__repr__() for b in self.binnings])

    def __add__(self, other) -> 'CompoundBinning':
        if isinstance(other, CompoundBinning):
            cb = CompoundBinning(self.binnings)
            cb.binnings.extend(other.binnings)
            cb.bins = concatenate([cb.bins, other.bins])
        elif isinstance(other, Binning):
            cb = CompoundBinning(self.binnings)
            cb.binnings.append(other)
            cb.bins = concatenate([cb.bins, other.bins])
        else:
            raise TypeError(f"Cannot concatenate CompoundBinning and {other.__class__.__name__}")
        return cb



#  ExoIris: fast, flexible, and easy exoplanet transmission spectroscopy in Python.
#  Copyright (C) 2024 Hannu Parviainen
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .exoiris import ExoIris, load_model
from .tsdata import TSData, TSDataSet
from .binning import Binning
from .ldtkld import LDTkLD

__version__ = '0.9.10'
